/*************************************************************************
  Expr        Calculates the value of a specified mathematical expression

  Author:     SemWare

  Date:       Mar 23, 1994
              Apr 18, 1996  Peter Birch: Added Exponentiation - **
              Mar  7, 1997  Peter Birch: Added paste-result ability
              Mar 2004  SEM: force grey-operator keys to
                self-insert - suggested by Ron Perrella.
              Jun 2005 SEM: Use commas with big numbers.

  Overview:

  This macro operates as a simple calculator, and handles decimal or
  hexadecimal numbers.  For decimal numbers, only whole numbers are
  accepted.

  Keys:       (none)

  Usage notes:

  This macro does not have any key assignments.  To use, simply select
  it from the Potpourri menu.

  The following operations are available:

            >      SHR      (bitwise shift right)
            <      SHL      (bitwise shift left)
            ~      COMP     (2's complement)
            %      MOD      (modulo division)
            &      AND      (bitwise AND)
            |      OR       (bitwise OR)
            ^      XOR      (bitwise eXclusive-OR)
            +      Addition
            -      Subtraction
            *      Multiplication
            **     Exponentiation--binary (Power)
            /      Division

  Precedence is as in SAL, but can be forced by using parenthesis.

  Hex number may be entered using C format, or ending with an 'h' or
  '$'.  Regardless, the first character _must_ be numeric.

  Copyright 1992-1995 SemWare Corporation.  All Rights Reserved Worldwide.

  Use, modification, and distribution of this SAL macro is encouraged by
  SemWare provided that this statement, including the above copyright
  notice, is not removed; and provided that no fee or other remuneration
  is received for distribution.  You may add your own copyright notice
  to cover new matter you add to the macro, but SemWare Corporation will
  neither support nor assume legal responsibility for any material added
  or any changes made to the macro.

*************************************************************************/

HelpDef ExprHelp
   "The following operators are available within an Expresson:"
   ""
   "Operator        Meaning                     Precedence"
   ""
   "    ()     Sub-expression                        1"
   ""
   "    ~      Bitwise 2's Complement                2"
   ""
   "    *      Multiplication                        3"
   "    **     Exponentiation                        3"
   "    /      Division                              3"
   "    %      Modulo Division                       3"
   "    ^      Bitwise eXclusive-OR                  3"
   "    >      Bitwise Shift Right                   3"
   "    <      Bitwise Shift Left                    3"
   "    &      Bitwise AND                           3"
   ""
   "    +      Addition                              4"
   "    -      Subtraction                           4"
   "    |      Bitwise OR                            4"
   ""
   "    Alt-P  Insert Result into current line        "
   ""
   "Hex number may be entered by beginning with '0x' or"
   "ending with an 'h' or '$'.  Regardless, the first"
   "character _must_ be numeric."
   ""
end

constant ins_DECIMAL        =   0x01,   // insert the sum in decimal
         ins_HEX            =   0x02    // insert the sum in hex

/*
   Work variables
*/

string  InputLine[80]
integer Col             // current column in InputLine
string  Token[15]       // currently parsed token from InputLine
integer symbol          // numeric "code" representing what token is
string  Ch[1]           // next Char from InputLine

integer finalResult
integer bufferId

integer hex             // is current symbol hex? used only by GetNextToken/isHex

forward integer proc SimpleExpression()

proc GetPrevChar()
    col = col - 1
    Ch = Upper(SubStr(inputline, col, 1))
end

proc GetNextChar()
    Ch = Upper(SubStr(inputline, col, 1))
    col = col + 1
end

integer proc isNumeric()
    return (Ch >= '0' and Ch <= '9')
end

integer proc isHexChar()
    if Ch >= 'A' and Ch <= 'F'
        hex = TRUE
        return (TRUE)
    endif
    return (FALSE)
end

// **************************************************************************
// Peter Birch  04/18/96
// my own cheesy pow() function
// overflows are easy with any exponent larger than 30
//
integer proc pow (integer base, integer exponent)
    integer iRetval, save_exponent, check

    save_exponent = exponent
    iRetval = base

    while exponent > 1
        iRetval = iRetval * base
        exponent = exponent - 1
    endwhile

    check = iRetval
    exponent = save_exponent
    while exponent > 1
        check = check / base
        exponent = exponent - 1
    endwhile

    if check <> base
        Warn("Overflow! ", base, " ** ", save_exponent, " : Result is too large")
    endif

    return (iRetval)
end

/*
   These are "magic" values that GetNextToken will store in the global variable
   symbol, to indicate what is the next item coming.
*/
constant
    symNumber           = 0,
    symAddition         = 1,
    symSubtraction      = 2,
    symMultiplication   = 3,
    symDivision         = 4,
    symLeftParen        = 5,
    symRightParen       = 6,
    symAND              = 7,
    symOR               = 8,
    symXOR              = 9,
    symCOMP             = 10,
    symMOD              = 11,
    symSHR              = 12,
    symSHL              = 13,
    symPower            = 14,
    symNot              = 15,
    symEndOfString      = -1,
    symError            = -2

// GetNextToken _must_ set hex to FALSE on startup and exit.
proc GetNextToken()
    while ch == " "                     //  Skip over any leading blanks
        GetNextChar()
    endwhile
    token = Ch
    case token
        when ">", ">>", "shr"   symbol = symSHR
        when "<", "<<", "shl"   symbol = symSHL
        when "~"                symbol = symCOMP
        when "%", "mod"         symbol = symMOD
        when "&"                symbol = symAND
        when "|"                symbol = symOR
        when "^"                symbol = symXOR
        when "+"                symbol = symAddition
        when "-"                symbol = symSubtraction
        when "*"
                GetNextChar()
                if (ch == '*')
                    symbol = symPower
                else
                    symbol = symMultiplication
                    GetPrevChar()
                endif

        when "/", "div"         symbol = symDivision
        when "("                symbol = symLeftParen
        when ")"                symbol = symRightParen
        when "!", "not"         symbol = symNot
        when ""                 symbol = symEndOfString
        otherwise     // if token is numeric, collect rest of digits
            hex = FALSE
            if not isNumeric() and not isHexChar()
                symbol = symError
            else
                symbol = symNumber          // Tag this as an integer value
                GetNextChar()
                if (Ch == "X")              // Allow c-style hex strings
                    hex = TRUE
                    GetNextChar()
                endif
                while isNumeric() or isHexChar()
                    token = token + ch
                    GetNextChar()
                endwhile
                if Ch == 'H' or Ch == '$'   // Allow assembly/Turbo Pascal hex strings
                    hex = TRUE
                    GetNextChar()
                elseif Ch == '#'
                    GetNextChar()
                    token = token + 'B'
                endif
                if hex
                    token = token + 'H'
                endif
            endif
            hex = FALSE
            return ()
    endcase
    GetNextChar()
end

//
//  Set up globals so SimpleExpression starts off on the right foot
//
proc SyntaxError()
    Warn("Expression syntax error")
    Halt
end

proc ParenError()
    Warn("Missing right parenthesis")
    Halt
end

integer proc Factor()
    integer result, base = 10

    case symbol
        when symNumber                  // found a number
            if token[Length(token)] == 'H'
                token = substr(token, 1, Length(token) - 1)
                base = 16
            elseif token[Length(token)] == 'B'
                token = substr(token, 1, Length(token) - 1)
                base = 2
            endif
            result = Val(token, base)            // get its value
            GetNextToken()                       // and skip to next
        when symLeftParen               // found a left paren
            GetNextToken()                       // skip over "("
            result = SimpleExpression()
            if  symbol <> symRightParen      // make sure the right paren exists
                ParenError()
            endif
            GetNextToken()                       // skip over ")"
        when symCOMP
            GetNextToken()
            result = ~Factor()
        when symNot
            GetNextToken()
            result = not Factor()
        when symSubtraction
            GetNextToken()
            result = - Factor()
        otherwise
            SyntaxError()                 // not number or paren--error
    endcase
    return  (result)
end

integer proc Term()
    integer result, sym

    result = Factor()
    while  (symbol == symMultiplication) or     // "*" sign, or
           (symbol == symDivision) or           // "/" sign...
           (symbol == symPower) or              // "**" sign...
           (symbol == symMOD) or                // "%" sign...
           (symbol == symXOR) or                // "^" sign...
           (symbol == symSHR) or                // ">" sign...
           (symbol == symSHL) or                // "<" sign...
           (symbol == symAND)                   // "&" sign...
        sym = symbol
        GetNextToken()                      // skip over math op
        case sym
            when symMultiplication
                result = result * Factor()    // multiplication
            when symPower
                result = pow(result, Factor())
            when symDivision
                result = result / Factor()    // division
            when symAND
                result = result & Factor()
            when symXOR
                result = result ^ Factor()
            when symMOD
                result = result mod Factor()
            when symSHR
                result = result shr Factor()
            when symSHL
                result = result shl Factor()
        endcase
    endwhile
    if symbol == symError
        SyntaxError()
    endif
    return (result)
end

integer proc SimpleExpression()
    integer result, sym

    result = Term()
    while  (symbol == symNumber     ) or       // another integer coming, or
           (symbol == symOR         ) or       // "|" sign, or
           (symbol == symAddition   ) or       // "+" sign, or
           (symbol == symSubtraction)          // "-" sign....
        sym = symbol
        if sym == symNumber            // if another integer came immediately
            sym = symAddition          // then default to adding
        else
            GetNextToken()             // else skip over the math op
        endif
        case sym
            when symAddition
                result = result + Term()
            when symSubtraction
                result = result - Term()
            when symOR
                result = result | Term()
        endcase
    endwhile
    if symbol == symError
        SyntaxError()
    endif
    return (result)
end

// Added by PLB 03/07/97
proc InsertExpr (integer flags)
    integer iPromptId

    iPromptId = getbufferId()
    gotoBufferId(bufferId)

    InsertText(iif(flags & ins_DECIMAL, str(finalResult, 10), str(finalResult,
               16) + "h"), _INSERT_)

    gotoBufferId(iPromptId)
end

Menu InsertExprMenu()
    title = "Insert Result at Cursor"

    "Insert Decimal Result" ,   InsertExpr(ins_DECIMAL)
    "Insert Hex Result"     ,   InsertExpr(ins_HEX)
end

keydef ExprHelpKey
    <F1>            QuickHelp(ExprHelp)
    <Alt p>         InsertExprMenu()
    <grey/>         SelfInsert()
    <grey*>         SelfInsert()
    <grey->         SelfInsert()
    <grey+>         SelfInsert()
end

proc PromptStartup()
    Enable(ExprHelpKey)
    UnHook(PromptStartup)
end

// Assumes input string has no leading zeroes...
// Returns string with formatting commas inserted
string proc InsertCommas(string s)
    string   ret[26] = s
    integer  idx = (length(s) mod 3) + 1
    if length(s) > 3
        while idx < length(ret)
            if idx > 1
                ret = InsStr(",",ret,idx)
                idx = idx + 4
            else
                idx = idx + 3
            endif
        endwhile
    endif
    return( ret)
end

proc Main()
    integer from_cmdline, hist
    bufferId = getBufferId()

    hist = GetFreeHistory("expr:main")
    InputLine = Trim(Query(MacroCmdLine))
    from_cmdline = Length(InputLine) > 0

    loop
        if not from_cmdline
            Hook(_PROMPT_STARTUP_, PromptStartup)
            if not Ask("Enter Expression: (<F1> for help)", InputLine, hist) or Length(InputLine) == 0
                break
            endif
        endif
        // Prime the pump...
        Col   = 1
        Token = ''
        GetNextChar()
        GetNextToken()

        finalResult = SimpleExpression()
        Message("The result is: ", InsertCommas(Str(finalResult)), " (", str(finalResult, 16), "h)", " (", str(finalResult, 2), "b)")
        if from_cmdline
            break
        endif
    endloop
end

